with PHYSICAL_UNITS_BASIC ; use PHYSICAL_UNITS_BASIC ;
with PHYSICAL_UNITS_MECHANICAL ; use PHYSICAL_UNITS_MECHANICAL ;
with PHYSICAL_UNITS_ELECTRICAL ; use PHYSICAL_UNITS_ELECTRICAL ;
with PHYSICAL_UNITS_OTHER ; use PHYSICAL_UNITS_OTHER ;

package PHYSICAL_CONSTANT is

--
--   There are a number of physical constants that are used in equations
-- to solve problems in physics. Errors may occur because the dimensionality
-- and/or units of the physical constant are not known. The table below
-- presents some physical constants with their typical symbol, dimension,
-- nominal value and unit of measure in the MKS system.
--
--   The Ada named munber is then given with the Ada type corresponding
-- to the required dimension and unit.
--
--PHYSICAL CONSTANT           SYMBOL  DIMENSION   MKS VALUE        UNIT
--_________________           ______  _________   _________        ____
--
--
--Pi, ratio of circumference   Pi      none        3.14159265       radians
--      to diameter
--
  PI : constant := 3.14159265 ;

--
--                                        3                             3
--air density, normal          rho     M/L         1.2929           Kg/m
--   conditions
--
  AIR_DENSITY : constant DENSITY_MKS := DIMENSION ( 1.2929 ) ;

--
--air molecule, mass           m       M           4.81E-26         Kg
--
  MASS_AIR_MOLECULE : constant MASS_MKS := DIMENSION ( 4.81E-26 ) ;

--
--air molecule,                w       M           0.028952         Kg/mole
-- kilogram molecular weight
--
  MOLECULAR_WEIGHT_AIR_MOLECULE : constant MASS_MKS := DIMENSION ( 0.028952 )
      ;

--
--                                         2                                2
--atmospheric pressure         A       M/LT        1.01325          newton/m
--
  ATMOSPHERIC_PRESSURE : constant PRESSURE_MKS := DIMENSION ( 1.01325 ) ;

--
--Avogadro's number            N       none        6.023E+23        molecules in
--  molecules in a kilogram mole                                          a mole
--
  AVOGADROS_NUMBER : constant := 6.02257E+23 ;

--
--                                       2  2                             o
--Boltzmann's constant         k       ML /T K     1.380E-23        joule/ K
--
  BOLTZMANNS : constant JOULE_PER_DEGREE_KELVIN := DIMENSION ( 1.380E-23 ) ;

--
--                                       2  2
--electron volt                e       ML /T       1.60210E-10      joule
--
  ELECTRON_VOLT : constant ENERGY_MKS := DIMENSION ( 1.60210E-10 ) ;

--
--                                       3  2 2                         2    -2
--electrostatic constant       k       ML /T Q     8.987E+9         nt m/coulomb
--  reciprocal permittivity                                          m/farad
--
  ELECTROSTATIC : constant METER_PER_FARAD := DIMENSION ( 8.987E+9 ) ;

--
--elementary charge            e       Q           1.6021892E-19    coulomb
--
  ELEMENTARY_CHARGE : constant CHARGE_COULOMB := DIMENSION ( 1.6021892E-19 ) ;

--
--electron mass                m       M           9.1066E-31       Kg
--                              e
--
  MASS_ELECTRON : constant MASS_MKS := DIMENSION ( 9.1066E-31 ) ;

--
--faraday                      f       L/T         9.648456E+4      coulomb/mole
--
  FARADAY : constant VELOCITY_MKS := DIMENSION ( 9.648456E+4 ) ;

--
--                                       2  2                             o
--gas constant of a mole       R       ML /T K     8.3144           joule/ K
--
  GAS : constant JOULE_PER_DEGREE_KELVIN := DIMENSION ( 8.3144 ) ;

--
--                                        2                              2
--gravity (earth)              g       L/T         9.80665          m/sec
--
  GRAVITY : constant ACCELERATION_MKS := DIMENSION ( 9.800665 ) ;

--
--
--hydrogen atom mass           m       M           1.6734E-27       Kg
--                              h
--
  MASS_HYDROGEN_ATOM : constant MASS_MKS := DIMENSION ( 1.6734E-27 ) ;

--
--hydrogen atom                w       M           1.0079E-3        Kg/mole
--  kilogram atomic weight
--
  MOLECULAR_WEIGHT_HYDROGEN_ATOM : constant MASS_MKS := DIMENSION ( 1.0079E-3
      ) ;

--
--                                       2   2
--impedance of free space      Z       ML /TQ      120Pi            ohm
--                              0
--
  IMPEDANCE_FREE_SPACE : constant RESISTANCE_OHM := DIMENSION ( 120.0 * PI ) ;

--
--mechanical equivalent        J       none        4186.05          joule/
--   of heat                                                          Kg calorie
--
  MECHANICAL_HEAT : constant := 4186.05 ;

--
--                                      2 2   3 
--permittivity (vacuum)        epsi    T Q /ML     8.854E-12        farad/meter
--                                 0
--
  PERMITTIVITY_VACUUM : constant PERMITTIVITY_FARAD_PER_METER := DIMENSION (
      8.854E-12 ) ;

--
--                                         2                        
--permeability (vacuum)        mu      ML/Q        4Pi E-7          henry/meter
--                               0
  PERMEABITITY_VACUUM : constant PERMEABILITY_HENRY_PER_METER := DIMENSION (
      4.0E-7 * PI ) ;

--                                       2                            
--Planck's constant            h       ML /T       6.624E-34        joule second
--
  PLANCKS : constant INERTIA_MKS := DIMENSION ( 6.624E-34 ) ;

--
--speed of light (vacuum)      c       L/T         2.99792458E+8    meter/second
--
  SPEED_OF_LIGHT : constant VELOCITY_MKS := DIMENSION ( 2.99792458E+8 ) ;

--
--speed of sound (air)         s       L/T         331.45           meter/second
--
  SPEED_OF_SOUND : constant VELOCITY_MKS := DIMENSION ( 331.45 ) ;

--
--                                      2   2                           2   2
--universal gravitational      G       L /MT       6.6720E-12       nt m /Kg
--    constant
--
  U_G_C : constant ACCELERATION_PER_KILOGRAM := DIMENSION ( 6.6720E-12 ) ;

--
end PHYSICAL_CONSTANT ;
